/**
* @file app_cluster_on_off.c
* @brief Common process - adapter the flash api
* @version 0.1
* @date 2021-10-22
*
* @copyright Copyright 2020-2021 Tuya Inc. All Rights Reserved.
*
*/
#include "tal_log.h"
#include "tal_memory.h"
#include "tal_system.h"
#include "tal_sw_timer.h"
#include "tal_attribute_rw.h"
#include "app_cluster_on_off.h"

#define TY_CMD_SET_COUNT_DOWN_CMD_ID 0xF0 ///<   tuya private cmd
#define COUNT_DOWN_MAX 86400

STATIC UCHAR_T count_down_report_onoff_flag = FALSE;
STATIC ULONG_T count_down_report_delay = 0;
extern TIMER_ID etimer_countdown;

/**
 * @note app on/off command handler
 * @param [in]{UCHAR_T} onoff
 * @param [in]{ZIGBEE_CMD_T} single or group transmit cmd
 * @return: none
 */
STATIC VOID_T __app_onoff_cmd_set_value_handler(ZIGBEE_CMD_T cmd_type, UCHAR_T onoff)
{
    TAL_SEND_QOS_E QOS = QOS_1;
    ULONG_T delay_ms = 0;
    OPERATE_RET ret;
    TAL_PR_DEBUG("Receive on/off: %d", onoff);

    if (cmd_type == ZIGBEE_CMD_GROUP)
    {
        delay_ms = 10000 + tal_system_get_random(1000);
        QOS = QOS_0;
    }
    app_onoff_report_value(QOS, delay_ms, onoff);
    //dont report onoff again in the count down handler
    count_down_report_onoff_flag = 0;
    count_down_report_delay = delay_ms;
    app_onoff_data_count_down_set(0);

    ret = app_light_ctrl_data_switch_set(onoff);
    if (ret != OPRT_OK)
    {
        return;
    }
    app_light_ctrl_proc();

    //update on/off attribute
    tal_zg_write_attribute(1,
                           CLUSTER_ON_OFF_CLUSTER_ID,
                           ATTR_ON_OFF_ATTRIBUTE_ID,
                           &onoff,
                           ATTR_BOOLEAN_ATTRIBUTE_TYPE);
}

/**
 * @brief 
 * 
 * @param Qos 
 * @param delay_ms 
 * @param onoff 
 * @return VOID_T 
 */
VOID_T app_onoff_report_value(TAL_SEND_QOS_E Qos, USHORT_T delay_ms, UINT8_T onoff)
{
    TAL_ZG_SEND_DATA_T send_data;

    tal_system_memset(&send_data, 0, sizeof(TAL_ZG_SEND_DATA_T));

    send_data.delay_time = delay_ms;
    send_data.random_time = 0;
    send_data.zcl_id = ZCL_ID_ONOFF;
    send_data.qos = Qos;
    send_data.direction = ZG_ZCL_DATA_SERVER_TO_CLIENT;
    send_data.frame_type = ZG_ZCL_FRAME_TYPE_GLOBAL;
    send_data.command_id = CMD_REPORT_ATTRIBUTES_COMMAND_ID;

    // unicast to gateway
    send_data.addr.mode = SEND_MODE_DEV;
    send_data.addr.type.dev.dst_addr = 0xfffc;
    send_data.addr.type.dev.dst_ep = 0x01;
    send_data.addr.type.dev.src_ep = TUYA_PRIMARY_ENDPOINT;
    send_data.addr.type.dev.cluster_id = CLUSTER_ON_OFF_CLUSTER_ID;

    send_data.data.zg.attr_sum = 1;
    send_data.data.zg.attr[0].attr_id = ATTR_ON_OFF_ATTRIBUTE_ID;
    send_data.data.zg.attr[0].type = ATTR_BOOLEAN_ATTRIBUTE_TYPE;
    send_data.data.zg.attr[0].size = 1;
    send_data.data.zg.attr[0].value[0] = onoff;
    tal_zg_send_data(&send_data, NULL, 2000);
    TAL_PR_DEBUG("Report onoff qos:%d VALUE %d", Qos, onoff);
}
/**
 * @note app_onoff_report_count_down_data
 * @param [in] {TAL_SEND_QOS_E} QOS 
 * @param [in] {USHORT_T} delay_ms
 * @param [in] {USHORT_T} countdown
 * @return: none
 */
VOID_T app_onoff_report_count_down_data(TAL_SEND_QOS_E Qos, USHORT_T delay_ms, ULONG_T countdown)
{
    TAL_ZG_SEND_DATA_T send_data;
    tal_system_memset(&send_data, 0, sizeof(TAL_ZG_SEND_DATA_T));

    send_data.delay_time = delay_ms;
    send_data.qos = Qos;
    send_data.zcl_id = ZCL_ID_COUNT_DOWN;
    send_data.direction = ZG_ZCL_DATA_SERVER_TO_CLIENT;
    send_data.command_id = CMD_REPORT_ATTRIBUTES_COMMAND_ID;
    send_data.addr.mode = SEND_MODE_DEV;
    send_data.addr.type.dev.cluster_id = CLUSTER_ON_OFF_CLUSTER_ID;
    send_data.addr.type.dev.src_ep = 1;
    send_data.addr.type.dev.dst_ep = 1;
    send_data.data.zg.attr_sum = 1;
    send_data.data.zg.attr[0].attr_id = TY_ATTR_COUNT_DOWN_ATTR_ID;
    send_data.data.zg.attr[0].type = ATTR_INT32U_ATTRIBUTE_TYPE;
    send_data.data.zg.attr[0].size = 4;
    send_data.data.zg.attr[0].value[0] = countdown;
    send_data.data.zg.attr[0].value[1] = countdown >> 8;
    send_data.data.zg.attr[0].value[2] = countdown >> 16;
    send_data.data.zg.attr[0].value[3] = countdown >> 24;

    //tal_zg_clear_send_data(ZG_CLEAR_ALL_ZCL_ID, &send_data.zcl_id);
    tal_zg_send_data(&send_data, NULL, 1000);
    TAL_PR_DEBUG("Report count down: %d", countdown);
}
/**
 * @note app_onoff_count_down_response
 * @param [in] {ULONG_T} remain_time_sec
 * @return: none
 */
VOID_T app_onoff_count_down_response(ULONG_T remain_time_sec)
{
    STATIC ULONG_T last_remain_sec = 0;
    UINT8_T onoff = 0;

    TAL_PR_DEBUG("%s", __FUNCTION__);
    if (last_remain_sec == remain_time_sec)
    {
        TAL_PR_DEBUG("last remian is equal to reamin!");
    }
    else
    {
        app_onoff_report_count_down_data(QOS_1, count_down_report_delay, (ULONG_T)remain_time_sec);
    }

    if (remain_time_sec == 0)
    {
        app_light_ctrl_data_switch_get(&onoff);
        TAL_PR_DEBUG("count_down_report_onoff_flag= %d", count_down_report_onoff_flag);
        if (count_down_report_onoff_flag == TRUE)
        {
            TAL_PR_DEBUG("count_down_report_delay %d", count_down_report_delay);
            app_onoff_report_value(QOS_0, count_down_report_delay, onoff);
            app_onoff_set_count_down_report_flag_time(FALSE, 1000);
            TAL_PR_DEBUG("report onoff %d", onoff);
        }
        tal_zg_write_attribute(1,
                               CLUSTER_ON_OFF_CLUSTER_ID,
                               ATTR_ON_OFF_ATTRIBUTE_ID,
                               &onoff,
                               ATTR_BOOLEAN_ATTRIBUTE_TYPE);
    }
    last_remain_sec = remain_time_sec;
}
/**
 * @note: app_onoff_count_down_timer_cb
 * @note: count down cnt 60s to report value
 * @param {none}
 * @return: none
 */
VOID_T app_onoff_count_down_timer_cb(TIMER_ID timer_id, VOID_T *arg)
{
    OPERATE_RET ret = -1;
    ULONG_T cnt = app_light_count_down_cnt_get();
    UINT8_T onoff = FALSE;
    if (cnt > 1)
    {
        cnt--;
        app_light_count_down_cnt_set(cnt);
        if ((cnt % 60) == 0)
        {
            app_onoff_count_down_response(cnt);
        }
    }
    else
    {
        app_light_count_down_cnt_set(0);
        app_light_ctrl_data_switch_get(&onoff);
        onoff = (onoff != FALSE) ? FALSE : TRUE;
        app_light_ctrl_data_switch_set(onoff);
        ret = app_light_ctrl_proc();
        if (ret != OPRT_OK)
        {
            TAL_PR_DEBUG("CountDown process error!");
        }
        app_onoff_count_down_response(0);
        return;
    }
    tal_sw_timer_start(etimer_countdown, 1000, TAL_TIMER_ONCE);
}
/**
 * @note: set light countdown value
 * @param {IN ULONG_T count_down_cnt -> unit:second}
 * @attention: countdown lave time will return with
 *              calling app_onoff_count_down_response function every minutes.
 *              switch status will return with calling
 *              light_ctrl_data_switch_response function when countdown active.
 * @return: OPERATE_RET -> OPRT_OK set countdown OK.
 */
OPERATE_RET app_onoff_data_count_down_set(ULONG_T count_down_cnt)
{
    if (count_down_cnt > COUNT_DOWN_MAX)
    {
        TAL_PR_DEBUG("Set count down value error!");
        return OPRT_INVALID_PARM;
    }
    app_light_count_down_cnt_set(count_down_cnt);
    if (count_down_cnt <= 0)
    {
        //cancel coutdown proc
        tal_sw_timer_stop(etimer_countdown);
        app_onoff_count_down_response(0);
        return OPRT_OK;
    }
    else
    {
        tal_sw_timer_start(etimer_countdown, 1000, TAL_TIMER_ONCE);
    }

    //upload countdown value
    app_onoff_count_down_response(count_down_cnt);

    return OPRT_OK;
}
/**
 * @note app_onoff_handle_count_down
 * @param [in]{UCHAR_T} onoff
 * @param [in]{ZIGBEE_CMD_T} cmd_type
 * @return: none
 */
VOID_T app_onoff_handle_count_down(ULONG_T delay_ms)
{
    UINT8_T pre_onoff;
    TAL_PR_DEBUG("%s", __FUNCTION__);
    app_light_ctrl_data_switch_get(&pre_onoff);
    app_light_ctrl_data_switch_set(1);
    if (pre_onoff != 1)
    {
        app_onoff_report_value(QOS_0, delay_ms, 1);
        count_down_report_onoff_flag = 0;
        count_down_report_delay = delay_ms;
        app_onoff_data_count_down_set(0);
    }
}
/**
 * @brief 
 * 
 * @param flag 
 * @param delay_ms 
 */
VOID_T app_onoff_set_count_down_report_flag_time(UINT8_T flag, ULONG_T delay_ms)
{
    count_down_report_onoff_flag = flag;
    count_down_report_delay = delay_ms;
}
/**
 * @brief handle zcl on/off cluster command
 * 
 * @param cmd 
 * @param payload 
 * @param payload_len 
 * @param cmd_type 
 * @return OPERATE_RET 
 */
OPERATE_RET app_onoff_cluster_handler(UCHAR_T cmd, UINT8_T *payload, UINT8_T payload_len, ZIGBEE_CMD_T cmd_type)
{
    UINT8_T onoff_value;
    ULONG_T count_down_data = 0;
    switch (cmd)
    {
    case CMD_OFF_COMMAND_ID:
    {
        __app_onoff_cmd_set_value_handler(cmd_type, FALSE);
        break;
    }
    case CMD_ON_COMMAND_ID:
    {
        __app_onoff_cmd_set_value_handler(cmd_type, TRUE);
        break;
    }
    case CMD_TOGGLE_COMMAND_ID:
    {
        app_light_ctrl_data_switch_get(&onoff_value);
        __app_onoff_cmd_set_value_handler(cmd_type, !onoff_value);
        break;
    }
    case TY_CMD_SET_COUNT_DOWN_CMD_ID:
    {
        //tuya Countdown
        count_down_data = (ULONG_T)payload[0] + ((ULONG_T)payload[1] << 8) + ((ULONG_T)payload[2] << 16) + ((ULONG_T)payload[3] << 24);
        app_onoff_set_count_down_report_flag_time(TRUE, 1000);
        app_onoff_data_count_down_set(count_down_data);
        break;
    }
    default:
        break;
    }
    return OPRT_OK;
}
